from pathlib import Path

import click
from icepack import Age, SSH
from icepack.model import Compression

from icebox import NAME, VERSION, human_readable_size
from icebox.box import Box


@click.group()
@click.option(
    '--config', '-c',
    help='Base configuration directory.',
    type=click.Path(file_okay=False))
@click.pass_context
def icebox(ctx, config):
    """Encrypting Cold Storage Client"""
    ctx.obj = {}
    if config is None:
        base_path = Path(click.get_app_dir(NAME))
    else:
        base_path = Path(config)
    base_path.mkdir(mode=0o700, parents=True, exist_ok=True)
    ctx.obj['base_path'] = base_path


@icebox.group()
@click.argument('box-name')
@click.pass_context
def init(ctx, box_name):
    """Create a new box."""
    base_path = ctx.obj['base_path']
    box = Box(base_path.joinpath(box_name))
    if box.exists():
        raise click.ClickException('Box already exists.')
    ctx.obj['box'] = box


@init.command('folder')
@click.argument(
    'folder-path',
    type=click.Path(exists=True, file_okay=False, writable=True))
@click.pass_context
def init_folder(ctx, folder_path):
    """Create a folder-backed box."""
    box = ctx.obj['box']
    box.config['backend'] = 'folder'
    box.config['folder_path'] = folder_path
    click.echo('Initializing box.')
    try:
        box.init(log=click.echo)
    except Exception as e:
        raise click.ClickException('Initialization failed. ({})'.format(e))
    click.echo(f'- Your encryption keys are in {box.path}')
    click.echo('- Make sure to protect and backup this directory!')
    click.echo('Box initialized.')


@init.command('s3')
@click.argument('bucket')
@click.option(
    '--class', '-c', 'storage_class',
    help='Storage class. (Default: DEEP_ARCHIVE)',
    type=click.Choice(['GLACIER', 'DEEP_ARCHIVE']),
    default='DEEP_ARCHIVE')
@click.option('--profile', '-p', help='AWS profile.', default='default')
@click.pass_context
def init_s3(ctx, bucket, storage_class, profile):
    """Create an Amazon S3-backed box."""
    box = ctx.obj['box']
    box.config['backend'] = 's3'
    box.config['bucket'] = bucket
    box.config['profile'] = profile
    box.config['storage_class'] = storage_class
    box.config['tier'] = 'Bulk'
    click.echo('Initializing box.')
    try:
        box.init(log=click.echo)
    except Exception as e:
        raise click.ClickException('Initialization failed. ({})'.format(e))
    click.echo(f'- Your encryption keys are in {box.path}')
    click.echo('- Make sure to protect and backup this directory!')
    click.echo('Box initialized.')


@init.command('webdav')
@click.argument('url')
@click.option(
    '--username', '-u',
    help='WebDAV username.',
    prompt='WebDAV username')
@click.password_option(
    '--password', '-p',
    help='WebDAV password.',
    prompt='WebDAV password')
@click.pass_context
def init_webdav(ctx, url, username, password):
    """Create a WebDAV-backed box."""
    box = ctx.obj['box']
    box.config['backend'] = 'webdav'
    box.config['url'] = url
    if username:
        box.config['username'] = username
    if password:
        box.config['password'] = password
    click.echo('Initializing box.')
    try:
        box.init(log=click.echo)
    except Exception as e:
        raise click.ClickException('Initialization failed. ({})'.format(e))
    click.echo(f'- Your encryption keys are in {box.path}')
    click.echo('- Make sure to protect and backup this directory!')
    click.echo('Box initialized.')


@icebox.command()
@click.argument('box-name')
@click.argument('source', type=click.Path(exists=True))
@click.option('--comment', help='Source comment.')
@click.option(
    '--compression', '-c',
    help=f'Compression for all files. (Default: {Compression.GZ})',
    type=click.Choice([c.value for c in Compression]),
    default=Compression.GZ)
@click.option(
    '--mode',
    help='Store file/directory modes.',
    is_flag=True)
@click.option(
    '--mtime',
    help='Store file/directory modification times.',
    is_flag=True)
@click.pass_context
def put(ctx, box_name, source, comment, compression, mode, mtime):
    """Store data in a box."""
    base_path = ctx.obj['base_path']
    box = Box(base_path.joinpath(box_name))
    if not box.exists():
        raise click.ClickException('Box not found.')
    src_path = Path(source)
    src_name = src_path.name
    if box.contains(src_name):
        raise click.ClickException('Source already exists in box.')
    click.echo(f'Storing {src_name} in box.')
    try:
        box.store(
            src_path,
            comment=comment,
            compression=compression,
            mode=mode,
            mtime=mtime,
            log=click.echo)
    except Exception as e:
        raise click.ClickException(f'Operation failed. ({e})')
    click.echo(f'Stored {src_name} in box.')


@icebox.command()
@click.argument('box-name')
@click.argument('source')
@click.option(
    '--destination', '-d',
    help='Destination directory. (Default: Current directory)',
    type=click.Path(exists=True, file_okay=False, writable=True),
    default='.')
@click.option(
    '--option', '-o',
    help='A key=value option for the backend operation.',
    multiple=True)
@click.option(
    '--mode',
    help='Restore file/directory modes.',
    is_flag=True)
@click.option(
    '--mtime',
    help='Restore file/directory modification times.',
    is_flag=True)
@click.pass_context
def get(ctx, box_name, source, destination, option, mode, mtime):
    """Retrieve data from a box."""
    base_path = ctx.obj['base_path']
    box = Box(base_path.joinpath(box_name))
    if not box.exists():
        raise click.ClickException('Box not found.')
    if not box.contains(source):
        raise click.ClickException('Source not found in box.')
    click.echo(f'Retrieving {source} from box.')
    dst_path = Path(destination)
    backend_options = dict(o.split('=') for o in option)
    try:
        box.retrieve(
            source,
            dst_path,
            backend_options,
            mode=mode,
            mtime=mtime,
            log=click.echo)
    except Exception as e:
        raise click.ClickException(f'Operation failed. ({e})')
    click.echo(f'Retrieved {source} from box.')


@icebox.command()
@click.argument('box-name')
@click.argument('source')
@click.pass_context
def delete(ctx, box_name, source):
    """Delete data from a box."""
    base_path = ctx.obj['base_path']
    box = Box(base_path.joinpath(box_name))
    if not box.exists():
        raise click.ClickException('Box not found.')
    if not box.contains(source):
        raise click.ClickException('Source not found in box.')
    try:
        box.delete(source)
    except Exception as e:
        raise click.ClickException(f'Operation failed. ({e})')
    click.echo(f'Deleted {source} from box.')


@icebox.command()
@click.argument('box-name')
@click.pass_context
def list(ctx, box_name):
    """List the data in a box."""
    base_path = ctx.obj['base_path']
    box = Box(base_path.joinpath(box_name))
    if not box.exists():
        raise click.ClickException('Box not found.')
    total_size = 0
    for source in box.sources():
        total_size += source.size
        size = human_readable_size(source.size)
        if source.comment:
            click.echo(f'{source.name} | {size} | {source.comment}')
        else:
            click.echo(f'{source.name} | {size}')
    click.echo(f'Total size: {human_readable_size(total_size)}')


@icebox.command()
@click.argument('box-name')
@click.option(
    '--option', '-o',
    help='A key=value option for the backend operation.',
    multiple=True)
@click.pass_context
def refresh(ctx, box_name, option):
    """Refresh local information for a box."""
    base_path = ctx.obj['base_path']
    box = Box(base_path.joinpath(box_name))
    if not box.exists():
        raise click.ClickException('Box not found.')
    click.echo('Refreshing box.')
    backend_options = dict(o.split('=') for o in option)
    try:
        box.refresh(backend_options, log=click.echo)
    except Exception as e:
        raise click.ClickException(f'Operation failed. ({e})')
    click.echo('Refreshed box.')


@icebox.command()
@click.option(
    '--dependencies', '-d',
    help='Check the dependencies.',
    is_flag=True)
@click.pass_context
def version(ctx, dependencies):
    """Show the version information."""
    click.echo(f'{NAME} {VERSION}')
    if not dependencies:
        return
    age_version, age_keygen = Age.version()
    if age_version:
        click.echo(f'✅ age found. (Version: {age_version})')
    else:
        click.echo(f'❌ age not found.')
    if age_keygen:
        click.echo(f'✅ age-keygen found.')
    else:
        click.echo(f'❌ age-keygen not found.')
    ssh_version, ssh_keygen = SSH.version()
    if ssh_version:
        click.echo(f'✅ ssh found. (Version: {ssh_version})')
    else:
        click.echo(f'❌ ssh not found.')
    if ssh_keygen:
        click.echo(f'✅ ssh-keygen found.')
    else:
        click.echo(f'❌ ssh-keygen not found.')



import boto3

from icepack.helper import File

from icebox.data import BackendFile, JobStatus


RESTORE_COMPLETE = 'ongoing-request="false"'
RESTORE_RUNNING = 'ongoing-request="true"'


class Backend():
    """Backend for Amazon S3-backed boxes."""

    def __init__(self, box_path, box_config):
        self.box_path = box_path
        self.box_config = box_config
        self.storage_class = box_config['storage_class']
        self.tier = box_config['tier']
        profile = box_config['profile']
        bucket = box_config['bucket']
        session = boto3.session.Session(profile_name=profile)
        self.bucket = session.resource('s3').Bucket(bucket)

    def box_init(self):
        """Optional box initialization at creation time."""
        list(self.bucket.objects.limit(count=1))  # Quick access test

    def store_data(self, src_path, name):
        """Store the data file as name, return a retrieval key."""
        return self._store(str(src_path), name, self.storage_class)

    def store_meta(self, src_path, name):
        """Store the metadata file as name, return a retrieval key."""
        return self._store(str(src_path), name, 'STANDARD')

    def retrieve_init(self, retrieval_key, options):
        """Initiate a retrieval job, return the job key."""
        obj = self.bucket.Object(retrieval_key)
        if obj.storage_class and obj.restore is None:
            tier = options.get('Tier', self.tier)
            request = {
                'Days': 1,
                'GlacierJobParameters': {'Tier': tier},
            }
            obj.restore_object(RestoreRequest=request)
        return retrieval_key

    def retrieve_status(self, job_key):
        """Return the JobStatus of the given job."""
        obj = self.bucket.Object(job_key)
        if obj.storage_class is None:
            return JobStatus.success
        if obj.restore is None:
            return JobStatus.failure
        if RESTORE_RUNNING in obj.restore:
            return JobStatus.running
        if RESTORE_COMPLETE in obj.restore:
            return JobStatus.success
        raise Exception(f'Unsupported restore state: {obj.restore}')

    def retrieve_finish(self, job_key):
        """Finish the job, return the temporary file's Path."""
        obj = self.bucket.Object(job_key)
        if obj.storage_class and \
                (obj.restore is None or RESTORE_COMPLETE not in obj.restore):
            raise Exception('Job was not successful.')
        tmp_path = File.mktemp()
        obj.download_file(str(tmp_path))
        return tmp_path

    def delete(self, retrieval_key):
        """Delete the data for the given retrieval key."""
        obj = self.bucket.Object(retrieval_key)
        obj.delete()

    def inventory_init(self):
        """Initiate an inventory job, return the job key."""
        return 'inventory'

    def inventory_status(self, job_key):
        """Return the JobStatus of the given job."""
        return JobStatus.success

    def inventory_finish(self, job_key):
        """Return a list of BackendFiles."""
        objects = self.bucket.objects.all()
        return [BackendFile(o.key, o.key, o.size) for o in objects]

    def _store(self, filename, name, storage_class):
        """Store filename as name, return a retrieval key."""
        args = {'StorageClass': storage_class}
        self.bucket.upload_file(filename, name, ExtraArgs=args)
        return name



NAME = 'icebox'
VERSION = '1.0.0'

SECRET_KEY = 'identity'  # nosec No secret
PUBLIC_KEY = 'identity.pub'
ALLOWED_SIGNERS = 'allowed_signers'


def human_readable_size(size, decimals=2):
    """Return size as a human-readable string, i.e. '42.23 MB'."""
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if abs(size) < 1024.0 or unit == 'TB':
            break
        size /= 1024.0
    return f'{size:.{decimals}f} {unit}'



from enum import Enum


class BackendFile():

    def __init__(self, key, name, size):
        self.key = key
        self.name = name
        self.size = size

    def __repr__(self):
        return f'<{self.name}|{self.key}|{self.size}>'


class JobStatus(Enum):
    """Status codes for long-running jobs."""

    running = 0
    success = 1
    failure = 2


class Source():
    """Local information on a stored source."""

    def __init__(self, name=None):
        self.name = name
        self.comment = None
        self.size = None
        self.data_key = None
        self.meta_key = None

    def __eq__(self, other):
        """Comparison method."""
        if not isinstance(other, Source):
            return False
        if self.name != other.name:
            return False
        if self.comment != other.comment:
            return False
        if self.size != other.size:
            return False
        if self.data_key != other.data_key:
            return False
        if self.meta_key != other.meta_key:
            return False
        return True



from shutil import rmtree
import sqlite3
import time
import uuid

import yaml

from icepack import IcepackReader, create_archive, extract_archive
from icepack.helper import File, SSH, Zip
from icepack.model import Compression
from icepack.meta import SECRET_KEY

from icebox import SECRET_KEY
from icebox.backend import get_backend
from icebox.data import JobStatus, Source


DATA_SUFFIX = '.data'
META_SUFFIX = '.meta'

SQL_SCHEMA_VERSION = '3'
SQL_CREATE_SOURCES = '''CREATE TABLE IF NOT EXISTS sources (
                        name text UNIQUE NOT NULL,
                        comment text,
                        size int NOT NULL,
                        data_key text NOT NULL,
                        meta_key text NOT NULL)
                        '''
SQL_CREATE_JOBS = '''CREATE TABLE IF NOT EXISTS jobs (
                     name text UNIQUE NOT NULL,
                     key text NOT NULL)
                     '''
SQL_CREATE_SETTINGS = '''CREATE TABLE IF NOT EXISTS settings (
                         key text UNIQUE NOT NULL,
                         value text)
                         '''

INVENTORY_JOB = '::inventory::'


class Box():
    """Box configuration and mappings."""

    def __init__(self, path):
        self.path = path
        self.config_file = self.path / 'config.yml'
        if self.exists():
            with open(self.config_file, 'r') as f:
                self.config = yaml.safe_load(f)
            self._db = SQLite(path)
        else:
            self.config = {}

    def exists(self):
        """Check if a configuration file for this box exists."""
        return self.config_file.exists()

    def init(self, log=lambda msg: None):
        """Initialize this box on creation."""
        if self.exists():
            raise Exception('Box already initialized.')
        self.path.mkdir(mode=0o700, parents=True, exist_ok=True)
        self._db = SQLite(self.path)
        if not (self.path / SECRET_KEY).exists():
            log('- Generating encryption keys.')
            SSH.keygen(self.path)
        backend = self._backend()
        backend.box_init()
        with open(self.config_file, 'w') as f:
            yaml.safe_dump(self.config, f, default_flow_style=False)

    def store(
            self,
            src_path,
            comment=None,
            compression=Compression.GZ,
            mode=False,
            mtime=False,
            log=lambda msg: None):
        """Encrypt the given source and store in backend."""
        if not self.exists():
            raise Exception('Box not found.')
        backend = self._backend()
        temp_path = File.mktemp(directory=True)
        data_path = temp_path / (src_path.name + '.zip')
        meta_path = temp_path / (src_path.name + META_SUFFIX)
        try:
            log('- Creating archive.')
            create_archive(
                src_path,
                data_path,
                self.path,
                comment=comment,
                compression=compression,
                mode=mode,
                mtime=mtime)
            _export_metadata(data_path, meta_path)
            data_name, meta_name = _backend_names()
            source = Source(src_path.name)
            log('- Transferring to backend.')
            source.comment = comment
            source.size = data_path.stat().st_size
            source.data_key = backend.store_data(data_path, data_name)
            source.meta_key = backend.store_meta(meta_path, meta_name)
            self._db.save_source(source)
        finally:
            rmtree(temp_path, ignore_errors=True)

    def retrieve(
            self,
            name,
            dst_path,
            backend_options,
            mode=False,
            mtime=False,
            log=lambda msg: None):
        """Retrieve source from the backend and decrypt."""
        if not self.exists():
            raise Exception('Box not found.')
        backend = self._backend()
        data_path = None
        try:
            source = self._db.load_source(name)
            # Get existing retrieval job or start a new one
            job = self._db.load_job(name)
            if job is None:
                log('- Initiating transfer from backend.')
                job = backend.retrieve_init(source.data_key, backend_options)
                self._db.save_job(name, job)
            # Wait until job is done
            status = backend.retrieve_status(job)
            if status == JobStatus.running:
                log('- Transfer from backend pending.')
            while status == JobStatus.running:
                time.sleep(60)
                status = backend.retrieve_status(job)
            if status == JobStatus.failure:
                self._db.delete_job(name)
                raise Exception('Transfer from backend failed.')
            # Download the data file
            log('- Transferring from backend.')
            data_path = backend.retrieve_finish(job)
            self._db.delete_job(name)
            # Decrypt original source
            log('- Extracting archive.')
            extract_archive(
                data_path,
                dst_path,
                self.path,
                mode=mode,
                mtime=mtime)
        finally:
            if data_path:
                data_path.unlink(missing_ok=True)

    def contains(self, source):
        """Return True if the source name exists in this box."""
        if not self.exists():
            raise Exception('Box not found.')
        return self._db.load_source(source) is not None

    def delete(self, source):
        """Delete encrypted data and metadata in the backend."""
        if not self.exists():
            raise Exception('Box not found.')
        backend = self._backend()
        src = self._db.load_source(source)
        backend.delete(src.data_key)
        backend.delete(src.meta_key)
        self._db.delete_source(source)

    def sources(self):
        """Return information about known sources."""
        if not self.exists():
            raise Exception('Box not found.')
        return self._db.load_sources()

    def refresh(self, backend_options, log=lambda msg: None):
        """Refresh local information from the backend."""
        if not self.exists():
            raise Exception('Box not found.')
        backend = self._backend()
        inventory_job = self._db.load_job(INVENTORY_JOB)
        if inventory_job is None:
            log('- Initiating inventory retrieval.')
            inventory_job = backend.inventory_init()
            self._db.save_job(INVENTORY_JOB, inventory_job)
        status = backend.inventory_status(inventory_job)
        if status == JobStatus.running:
            log('- Inventory retrieval pending.')
        while status == JobStatus.running:
            time.sleep(60)
            status = backend.inventory_status(inventory_job)
        if status == JobStatus.failure:
            self._db.delete_job(INVENTORY_JOB)
            raise Exception('Inventory retrieval failed.')
        log('- Inventory retrieval finished.')

        sources = self.sources()
        inventory = backend.inventory_finish(inventory_job)
        check = InventoryCheck(sources, inventory)

        jobs = {}
        for meta, data in check.importable.items():
            job = self._db.load_job(meta.key)
            if job is None:
                log(f'- Initiating metadata retrieval: {meta.key}')
                job = backend.retrieve_init(meta.key, backend_options)
                self._db.save_job(meta.key, job)
            jobs[job] = meta
        while jobs:
            finished = []
            for job, meta in jobs.items():
                status = backend.retrieve_status(job)
                if status != JobStatus.running:
                    self._db.delete_job(meta.key)
                if status == JobStatus.failure:
                    raise Exception(f'Metadata retrieval failed: {meta.key})')
                elif status == JobStatus.success:
                    meta_path = backend.retrieve_finish(job)
                    with IcepackReader(meta_path, self.path) as archive:
                        name = archive.metadata.archive_name
                    if name.endswith('.zip'):
                        name = name[:-4]
                    data = check.importable[meta]
                    src = Source(name)
                    src.comment = archive.metadata.comment
                    src.size = data.size
                    src.data_key = data.key
                    src.meta_key = meta.key
                    meta_path.unlink()
                    if not self.contains(src.name):
                        self._db.save_source(src)
                        log(f'- Added {src.name} = {meta.key}')
                    else:
                        log(f'- Ignoring duplicate {src.name} = {meta.key}')
                    finished.append(job)
            jobs = {j: k for j, k in jobs.items() if j not in finished}
            if jobs:
                log(f'- Metadata retrievals pending: {len(jobs)}')
                time.sleep(60)
        self._db.delete_job(INVENTORY_JOB)

        if check.broken_sources:
            log('- Sources with missing backend files:')
            for name, missing in check.broken_sources.items():
                log(f'  {name}: {" / ".join(missing)}')

        if check.broken_backend:
            log('- Dangling backend files:')
            for f in check.broken_backend:
                log(f'  {f.key}')

    def _backend(self):
        """Return a backend instance for this box."""
        return get_backend(self.config['backend'], self.path, self.config)


class InventoryCheck():
    """Find broken sources and broken/importable backend files."""

    def __init__(self, sources, inventory):
        backend_keys = {f.key: f for f in inventory}
        # Check sources
        self.broken_sources = {}
        for s in sources:
            data = backend_keys.pop(s.data_key, None)
            meta = backend_keys.pop(s.meta_key, None)
            if data and meta:
                continue
            missing = []
            if not data:
                missing.append(s.data_key)
            if not meta:
                missing.append(s.meta_key)
            self.broken_sources[s.name] = missing
        # Check backend files
        self.importable = {}
        self.broken_backend = []
        inventory = backend_keys.values()
        backend_data = self._map_by_name_suffix(inventory, DATA_SUFFIX)
        backend_meta = self._map_by_name_suffix(inventory, META_SUFFIX)
        for name, data in backend_data.items():
            meta = backend_meta.pop(_sibling(name), None)
            if meta:
                self.importable[meta] = data
            else:
                self.broken_backend.append(data)
        self.broken_backend.extend(backend_meta.values())

    @staticmethod
    def _map_by_name_suffix(inventory, suffix):
        """Map name to BackendFile by name suffix."""
        return {f.name: f for f in inventory if f.name.endswith(suffix)}


class SQLite():
    """SQLite wrapper for source and job state."""

    def __init__(self, path):
        db_path = path.joinpath('box.db')
        self._conn = sqlite3.connect(str(db_path))
        self._ensure_tables()

    def load_source(self, name):
        """Return a Source by name."""
        stmt = '''SELECT name, comment, size, data_key, meta_key
                  FROM sources
                  WHERE name=?
                  '''
        c = self._conn.execute(stmt, (name,))
        r = c.fetchone()
        c.close()
        if r is not None:
            return self._to_source(r)
        else:
            return None

    def load_sources(self):
        """Return an iterable for all Sources."""
        stmt = '''SELECT name, comment, size, data_key, meta_key
                  FROM sources
                  ORDER BY name COLLATE NOCASE
                  '''
        c = self._conn.execute(stmt)

        def generator():
            for r in c:
                yield self._to_source(r)
            c.close()

        return generator()

    def save_source(self, source):
        """Save a Source."""
        stmt = '''INSERT INTO sources
                  (name, comment, size, data_key, meta_key)
                  VALUES (?, ?, ?, ?, ?)
                  '''
        values = (
            source.name,
            source.comment,
            source.size,
            source.data_key,
            source.meta_key
        )
        with self._conn:
            self._conn.execute(stmt, values)

    def delete_source(self, name):
        """Delete a Source by name."""
        stmt = 'DELETE FROM sources WHERE name=?'
        with self._conn:
            self._conn.execute(stmt, (name,))

    def load_job(self, name):
        """Return the job key for the given name."""
        stmt = 'SELECT key FROM jobs WHERE name=?'
        c = self._conn.execute(stmt, (name,))
        r = c.fetchone()
        c.close()
        if r is not None:
            return r[0]
        else:
            return None

    def save_job(self, name, key):
        """Save the job key for the given name."""
        stmt = 'INSERT INTO jobs (name, key) VALUES (?, ?)'
        values = (name, key)
        with self._conn:
            self._conn.execute(stmt, values)

    def delete_job(self, name):
        """Delete the job key for the given name."""
        stmt = 'DELETE FROM jobs WHERE name=?'
        with self._conn:
            self._conn.execute(stmt, (name,))

    def _ensure_tables(self):
        """Ensure all tables exist and are up to date."""
        with self._conn:
            self._conn.execute(SQL_CREATE_SETTINGS)
        stmt = 'SELECT value FROM settings WHERE key="schema"'
        c = self._conn.execute(stmt)
        r = c.fetchone()
        c.close()
        if r is None:
            schema = None
        else:
            schema = r[0]
        if schema is None:
            self._create_tables()
        elif schema == SQL_SCHEMA_VERSION:
            pass
        else:
            raise Exception('Unsupported schema version: ' + str(schema))

    def _create_tables(self):
        """Create all tables from scratch."""
        schema = 'INSERT INTO settings VALUES ("schema", ?)'
        with self._conn:
            self._conn.execute(SQL_CREATE_SOURCES)
            self._conn.execute(SQL_CREATE_JOBS)
            self._conn.execute(schema, SQL_SCHEMA_VERSION)

    @staticmethod
    def _to_source(row):
        """Return a Source for the given row."""
        source = Source()
        source.name = row[0]
        source.comment = row[1]
        source.size = row[2]
        source.data_key = row[3]
        source.meta_key = row[4]
        return source


def _backend_names():
    """Return a tuple of names for encrypted data and metadata files."""
    basename = str(uuid.uuid4())
    data_name = basename + DATA_SUFFIX
    meta_name = basename + META_SUFFIX
    return data_name, meta_name


def _export_metadata(src_path, dst_path):
    """Export metadata from src_path to dst_path."""
    with Zip(src_path) as src:
        meta_path, sig_path = src.extract_metadata()
        with Zip(dst_path, 'w') as dst:
            dst.add_metadata(meta_path, sig_path)


def _sibling(name):
    """Return the metadata name for a data file and vice versa."""
    if name.endswith(DATA_SUFFIX):
        return name[:-len(DATA_SUFFIX)] + META_SUFFIX
    elif name.endswith(META_SUFFIX):
        return name[:-len(META_SUFFIX)] + DATA_SUFFIX
    else:
        raise Exception(f'Invalid name: {name}')



from webdav4.client import Client

from icepack.helper import File

from icebox.data import BackendFile, JobStatus


class Backend():
    """Backend for WebDAV-backed boxes."""

    def __init__(self, box_path, box_config):
        self.box_path = box_path
        self.box_config = box_config
        url = box_config['url']
        username = box_config.get('username')
        password = box_config.get('password')
        if not username:
            raise Exception('No username specified.')
        if not password:
            raise Exception('No password specified.')
        self.client = Client(url, auth=(username, password))

    def box_init(self):
        """Optional box initialization at creation time."""
        self.client.exists('access-test')  # Quick access test

    def store_data(self, src_path, name):
        """Store the data file as name, return a retrieval key."""
        self.client.upload_file(src_path, name)
        return name

    def store_meta(self, src_path, name):
        """Store the metadata file as name, return a retrieval key."""
        self.client.upload_file(src_path, name)
        return name

    def retrieve_init(self, retrieval_key, options):
        """Initiate a retrieval job, return the job key."""
        return retrieval_key

    def retrieve_status(self, job_key):
        """Return the JobStatus of the given job."""
        exists = self.client.exists(job_key)
        if exists:
            return JobStatus.success
        else:
            return JobStatus.failure

    def retrieve_finish(self, job_key):
        """Finish the job, return the temporary file's Path."""
        tmp_path = File.mktemp()
        self.client.download_file(job_key, tmp_path)
        return tmp_path

    def delete(self, retrieval_key):
        """Delete the data for the given retrieval key."""
        self.client.remove(retrieval_key)

    def inventory_init(self):
        """Initiate an inventory job, return the job key."""
        return 'inventory'

    def inventory_status(self, job_key):
        """Return the JobStatus of the given job."""
        return JobStatus.success

    def inventory_finish(self, job_key):
        """Return a list of BackendFiles."""
        result = []
        for c in self.client.ls('', detail=True):
            if c['type'] != 'file':
                continue
            file = BackendFile(c['name'], c['name'], c['content_length'])
            result.append(file)
        return result



from pathlib import Path
import shutil

from icepack.helper import File

from icebox.data import BackendFile, JobStatus


class Backend():
    """Backend for folder-backed boxes."""

    def __init__(self, box_path, box_config):
        self.box_path = box_path
        self.box_config = box_config
        self.folder_path = Path(box_config['folder_path'])

    def box_init(self):
        """Optional box initialization at creation time."""
        if not self.folder_path.exists():
            raise Exception(f'No such folder: {self.folder_path}')

    def store_data(self, src_path, name):
        """Store the data file as name, return a retrieval key."""
        return self._store(src_path, name)

    def store_meta(self, src_path, name):
        """Store the metadata file as name, return a retrieval key."""
        return self._store(src_path, name)

    def retrieve_init(self, retrieval_key, options):
        """Initiate a retrieval job, return the job key."""
        return retrieval_key

    def retrieve_status(self, job_key):
        """Return the JobStatus of the given job."""
        file_path = self.folder_path.joinpath(job_key)
        if file_path.exists():
            return JobStatus.success
        else:
            return JobStatus.failure

    def retrieve_finish(self, job_key):
        """Finish the job, return the temporary file's Path."""
        file_path = self.folder_path.joinpath(job_key)
        tmp_path = File.mktemp()
        shutil.copy(str(file_path), str(tmp_path))
        return tmp_path

    def delete(self, retrieval_key):
        """Delete the data for the given retrieval key."""
        file_path = self.folder_path.joinpath(retrieval_key)
        file_path.unlink(missing_ok=True)

    def inventory_init(self):
        """Initiate an inventory job, return the job key."""
        return 'inventory'

    def inventory_status(self, job_key):
        """Return the JobStatus of the given job."""
        return JobStatus.success

    def inventory_finish(self, job_key):
        """Return a list of BackendFiles."""
        result = []
        for c in self.folder_path.iterdir():
            if not c.is_file():
                continue
            result.append(BackendFile(c.name, c.name, c.stat().st_size))
        return result

    def _store(self, src_path, name):
        """Store the file at src_path as name, return a retrieval key."""
        dst_path = self.folder_path.joinpath(name)
        shutil.copy(str(src_path), str(dst_path))
        return name
